"""Additional checks and filters for problems."""
from __future__ import annotations
import ast
import inspect
from typing import Any, Callable, Iterable, Optional, TypeVar, Union
from unittest import TestCase

__all__ = ("Site", "Disallow")

Output = TypeVar("Output")

Site = tuple[str, int]


class Disallow:
    """A list of items to disallow in code.

    Attributes
    ----------
    functions : list[str]
        The names of functions which the student should not be able to call.
    binops : list[type]
        The types of binary operations wihch the student should not be able to use.
        E.x., to forbid floating-point division, use `ast.Div`. See
        `here <https://docs.python.org/3/library/ast.html#ast.BinOp>`_ for a list.
    nodes : list[type]
        The types of any ast nodes wihch the student should not be able to use.
        E.x., to forbid for loops, use `ast.For`. See
        `the docs <https://docs.python.org/3/library/ast.html#node-classes>`_ for a
        list.

    Examples
    --------
    To disallow the built-in `map` function: `Disallow(functions=["map"])`.

    To disallow the built-in `str.map` function: `Disallow(functions=["count"])`.
    Note that for class method names, you just use the name of the function.

    Note that there is no way to disallow `+=` without also disallowing `+` with this
    API.
    """

    def __init__(
        self,
        functions: Optional[list[str]] = None,
        binops: Optional[list[type]] = None,
        nodes: Optional[list[type]] = None,
    ):
        self._functions = functions or []
        self._binops = binops or []
        self._nodes = nodes or []

    def to_test(
        self,
    ) -> Callable[[TestCase, Callable[..., Output], Callable[..., Output]], None]:
        """Generate a test method suitable for `aga_override_test` of `test_case`.

        You can pass the output of this method directly to `aga_override_test`.

        You can also use the lower-level methods `search_on_object` or `search_on_src`
        if you want to generate your own error message.
        """

        def inner(
            case: TestCase,
            _: Callable[..., Output],
            student: Callable[..., Output],
        ) -> None:
            msg = ""
            for site in self.search_on_object(student):
                if msg == "":
                    msg = "Looks like use you used some disallowed constructs:\n"

                msg += f"  - {site[0]} on line {site[1]}\n"

            case.assertEqual(msg, "", msg)

        return inner

    def search_on_object(self, obj: Any) -> Iterable[Site]:
        """Search for disallowed AST objects in a python object."""
        yield from self.search_on_src(inspect.getsource(obj))

    def search_on_src(
        self,
        src: str,
    ) -> Iterable[Site]:
        """Search for disallowed AST objects in a source string."""
        # Walk through each AST node (with no guarantee of order)
        for node in ast.walk(ast.parse(src)):
            # handle function calls
            if self._functions != [] and isinstance(node, ast.Call):
                use = _check_function(node.func, self._functions)  # type: ignore
                if use is not None:
                    yield use

            # handle binops
            if self._binops != [] and isinstance(node, (ast.BinOp, ast.AugAssign)):
                use = _check_binop(node, self._binops)
                if use is not None:
                    yield use

            # handle other objects
            if self._nodes != []:
                use = _check_ast_node(node, self._nodes)
                if use is not None:
                    yield use


def _check_function(call: ast.Call, functions: list[str]) -> Optional[Site]:
    """Check whether call is in functions."""
    # get the function name
    name = None
    if isinstance(call, ast.Name):
        name = call.id
    elif isinstance(call, ast.Attribute):
        name = call.attr

    if name is not None and name in functions:
        return (name, call.lineno)

    return None


def _check_binop(
    op_node: Union[ast.BinOp, ast.AugAssign], binops: list[type]
) -> Optional[Site]:
    """Check whether the binop is disallowed."""
    binop = op_node.op
    if type(binop) in binops:
        name = type(binop).__name__
        return (name, op_node.lineno)

    return None


def _check_ast_node(node: Any, nodes: list[type]) -> Optional[Site]:
    """Check whether an arbitrary AST node is disallowed."""
    if type(node) in nodes:
        name = type(node).__name__
        return (name, node.lineno)

    return None



"""Test Suite for the aga.core module."""
from __future__ import annotations

from copy import deepcopy
from dataclasses import dataclass
from datetime import timedelta
from types import FunctionType
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    Sequence,
    Tuple,
    TypeVar,
    Iterable,
)
from unittest import TestCase, TestSuite
from unittest.mock import patch

from .parameter import _TestParam, AgaKeywordContainer
from .utils import CaptureOut, initializer, Initializer
from ..config import AgaConfig, AgaTestConfig
from ..score import Prize, ScoredPrize, ScoreInfo, compute_scores
from ..util import text_diff

__all__ = ["TestMetadata", "SubmissionMetadata", "AgaTestCase", "AgaTestSuite"]

Output = TypeVar("Output")

_NEW_LINE = "\n"


@dataclass(frozen=True)
class TestMetadata:
    """Stores metadata about a specific test case."""

    max_score: float
    config: AgaTestConfig
    check_stdout: bool
    mock_input: bool
    hidden: bool = False


@dataclass(frozen=True)
class SubmissionMetadata:
    """Metadata for testing a submission, collected from the frontend.

    Attributes
    ----------
    total_score : float
        The problem's total score.
    time_since-due : timedelta
        The delta _from_ the due date _to_ the submission date, i.e. it's negative if
        the problem was submitted before the due date.
    previous_submissions : int
        The number of previous submissions.

    """

    total_score: float
    time_since_due: timedelta
    previous_submissions: int

    def is_on_time(self) -> bool:
        """Return true of the submission was on time."""
        return self.time_since_due <= timedelta()


class AgaTestCase(TestCase):
    """A `TestCase` which runs a single test of a `Problem`."""

    def __init__(
        self,
        test_input: "_TestInputs[Output]",
        golden: Callable[..., Output],
        under_test: Callable[..., Output],
        metadata: TestMetadata,
    ) -> None:
        super().__init__()
        self._test_input = test_input
        self._golden = golden
        self._under_test = under_test
        self._metadata = metadata

    @property
    def metadata(self) -> TestMetadata:
        """Get the test's metadata."""
        return self._metadata

    @property
    def test_input(self) -> _TestInputs[Output]:
        """Get the test input."""
        return self._test_input

    # pylint: disable=invalid-name
    # camelCase is required by unittest
    def runTest(self) -> None:
        """Run the test case."""
        self._test_input.check(self._golden, self._under_test, self._metadata)

    # pylint: disable=invalid-name
    # camelCase is required by unittest
    def shortDescription(self) -> str:
        """Dynamically generate the test name.

        This method is called by unittest.
        """
        return self.name

    @property
    def name(self) -> str:
        """Format the name of the test case."""
        return (
            self._test_input.aga_kwargs.name
            or self._metadata.config.name_fmt.format(
                args=self._test_input.param.args_repr(self._metadata.config.name_sep),
                kwargs=self._test_input.param.kwargs_repr(
                    self._metadata.config.name_sep
                ),
                sep=self._test_input.param.sep_repr(self._metadata.config.name_sep),
            )
        )

    @property
    def description(self) -> str | None:
        """Get the problem's description."""
        return self._test_input.aga_kwargs.description


class AgaTestSuite(TestSuite):
    """A thin wrapper around TestSuite that store a config."""

    def __init__(self, config: AgaConfig, tests: list[AgaTestCase]):
        super().__init__(tests)
        self.config = config


def generate_custom_input(input_list: Iterable[str]) -> Callable[[Any], str]:
    """Generate a custom input function for a test case."""
    _iterator = iter(input_list)

    def _custom_input(*args: Any) -> str:
        print(*args)
        return next(_iterator)

    return _custom_input


class _TestInputs(TestCase, Generic[Output]):
    """A single set of test inputs for a problem.

    These will be run against the golden solution and the function under test; their
    outputs will be compared, and a unittest failure raised if they differ.
    """

    # this tells unittest not to print their default assertion error messages
    longMessage = False

    def __init__(
        self,
        aga_param: _TestParam,
        mock_input: bool,
    ) -> None:
        super().__init__()
        self._mock_input = mock_input
        self._param: _TestParam = aga_param
        self.score_info = ScoreInfo(
            self.aga_kwargs.weight, self.aga_kwargs.value, self.aga_kwargs.extra_credit
        )

    @property
    def args(self) -> Tuple[Any, ...]:
        """Get the positional arguments for the test case."""
        return self._param.args

    @property
    def kwargs(self) -> Dict[str, Any]:
        """Get the keyword arguments for the test case."""
        return self._param.kwargs

    @property
    def param(self) -> _TestParam:
        """Get the test parameter."""
        return self._param

    @property
    def aga_kwargs(self) -> AgaKeywordContainer:
        """Get the keyword arguments for the test case."""
        return self._param

    @property
    def mock_input(self) -> bool:
        """Get the mock input of the test case."""
        return self._mock_input

    @property
    def description(self) -> str | None:
        """Get the description of the test case."""
        return self.aga_kwargs.description

    @description.setter
    def description(self, desc: str | None) -> None:
        """Set the description of the test case."""
        self.aga_kwargs.description = desc

    @property
    def name(self) -> str | None:
        """Get the description of the test case."""
        return self.aga_kwargs.name

    @name.setter
    def name(self, name: str | None) -> None:
        """Set the description of the test case."""
        self.aga_kwargs.name = name

    def _eval_mock_input(
        self, answer: Callable[..., Output], check_output: bool = False
    ) -> Tuple[str | None, Output]:
        """Evaluate func on the arguments."""
        # deepcopy in case the student submission mutates arguments; we don't want it to
        # mess with our copy of the arguments
        with CaptureOut(check_output) as stdout, patch(
            "builtins.input", generate_custom_input(deepcopy(self.args))
        ):
            result = answer()

        return stdout.value, result

    def _test_mock_input(
        self,
        golden: Callable[..., Output],
        under_test: Callable[..., Output],
        eq_fn: Callable[[_TestInputs[Output], Output, Output, TestMetadata, str], bool],
        metadata: TestMetadata,
    ) -> None:
        golden_stdout, golden_result = self._eval_mock_input(
            golden, metadata.check_stdout
        )

        under_test_stdout, under_test_result = self._eval_mock_input(
            under_test, metadata.check_stdout
        )

        eq_fn(
            self,
            golden_result,
            under_test_result,
            metadata,
            metadata.config.failure_msg,
        )

        self._assert_output_eq(golden_stdout, under_test_stdout, metadata)

    def _eval_regular(
        self, answer: Callable[..., Output], capture_output: bool
    ) -> Tuple[str | None, Output]:
        """Evaluate func on the arguments."""
        with CaptureOut(capture_output) as stdout:
            result = answer(*deepcopy(self.args), **deepcopy(self.kwargs))

        return stdout.value, result

    def _test_regular(
        self,
        golden: Callable[..., Output],
        under_test: Callable[..., Output],
        eq_fn: Callable[[_TestInputs[Output], Output, Output, TestMetadata, str], bool],
        metadata: TestMetadata,
    ) -> None:
        golden_stdout, golden_result = self._eval_regular(golden, metadata.check_stdout)
        under_test_stdout, under_test_result = self._eval_regular(
            under_test, metadata.check_stdout
        )

        eq_fn(
            self,
            golden_result,
            under_test_result,
            metadata,
            metadata.config.failure_msg,
        )

        self._assert_output_eq(golden_stdout, under_test_stdout, metadata)

    def _eval_pipeline(
        self, answer: Callable[..., Output], check_output: bool
    ) -> Tuple[str | None, Sequence[Any]]:
        self.description = ""
        temp_name = initializer.DEFAULT_NEW_INSTANCE_NAME
        if isinstance(answer, (type, FunctionType)):
            temp_name = answer.__name__

        with CaptureOut(check_output) as stdout:
            results = [None]

            if len(self.args) > 0:
                pipeline_processes = self.args

                for process in pipeline_processes:
                    if isinstance(process, Initializer):
                        answer = process(answer)
                        results.append(None)
                        self.description += (
                            f"{process.DEFAULT_NEW_INSTANCE_NAME} = "
                            f"{temp_name}{process}\n"
                        )
                        temp_name = process.DEFAULT_NEW_INSTANCE_NAME
                    else:
                        results.append(process(answer, results[-1]))
                        self.description += f"{temp_name}{process}\n"

        return stdout.value, results

    def _test_pipeline(
        self,
        golden: Callable[..., Output],
        under_test: Callable[..., Output],
        eq_fn: Callable[[_TestInputs[Output], Output, Output, TestMetadata, str], bool],
        metadata: TestMetadata,
    ) -> None:
        golden_stdout, golden_result = self._eval_pipeline(
            golden, metadata.check_stdout
        )
        under_test_stdout, under_test_result = self._eval_pipeline(
            under_test, metadata.check_stdout
        )

        if len(under_test_result) != len(golden_result):
            raise AssertionError(
                f"Expected {len(golden_result)} results, "
                f"but got {len(under_test_result)}"
            )

        for i, (golden_res, under_test_res) in enumerate(
            zip(golden_result[1:], under_test_result[1:])
        ):
            eq_fn(
                self,
                golden_res,
                under_test_res,
                metadata,
                f"The first diff happens in {i}th process (`{self.args[i]}`) "
                f"of the pipeline.\n"
                "The output is {output} and the expected is {expected}\n"
                f"Everything before the {i}th process (`{self.args[i]}`) "
                f"of the pipeline is correct:\n"
                f"{_NEW_LINE.join(map(str, self.args[:i + 1]))}\n\n"
                f"{metadata.config.failure_msg}",
            )

        self._assert_output_eq(golden_stdout, under_test_stdout, metadata)

    def _pick_test_fn(
        self,
    ) -> Callable[
        [
            Callable[..., Output],
            Callable[..., Output],
            Callable[[_TestInputs[Output], Output, Output, TestMetadata, str], bool],
            TestMetadata,
        ],
        None,
    ]:
        if self.mock_input:
            test_fn = self._test_mock_input
        elif self.aga_kwargs.is_pipeline:
            test_fn = self._test_pipeline
        else:
            test_fn = self._test_regular

        return test_fn

    def _pick_eval_fn(
        self,
    ) -> Callable[[Callable[..., Output], bool], Tuple[str, Any]]:
        if self.mock_input:
            test_fn: Any = self._eval_mock_input
        elif self.aga_kwargs.is_pipeline:
            test_fn = self._eval_pipeline
        else:
            test_fn = self._eval_regular

        return test_fn  # type: ignore

    def check(
        self,
        golden: Callable[..., Output],
        under_test: Callable[..., Output],
        metadata: TestMetadata,
    ) -> None:
        """Assert that `golden` and `under_test` give the same output.

        This method relies on `TestCase.assertEqual`, which will raise a unittest error
        if they differ. This means it can be plugged into any unittest TestCase as a
        helper method.
        """
        if self.aga_kwargs.override_test:
            self.aga_kwargs.override_test(self, golden, under_test)
            return

        if self.aga_kwargs.override_check:
            check = self.aga_kwargs.override_check
        else:
            check = type(self)._assert_eq

        test_fn = self._pick_test_fn()

        test_fn(golden, under_test, check, metadata)

    def _assert_output_eq(
        self, expected: str | None, got: str | None, metadata: TestMetadata
    ) -> None:
        self._assert_eq(
            expected or "",
            got or "",
            metadata,
            metadata.config.stdout_differ_msg,
        )

    def _assert_eq(
        self,
        expected: Any,
        got: Any,
        metadata: TestMetadata,
        msg_format: str,
    ) -> None:
        """Assert that expected equals got, formatting `msg_format` if not."""
        # we can only diff strings
        if isinstance(expected, str) and isinstance(got, str):
            diff_explanation = metadata.config.diff_explanation_msg
            diff = text_diff(got, expected)
        else:
            diff_explanation = ""
            diff = ""

        if isinstance(expected, float) and isinstance(got, (float, int)):
            comparator = self.assertAlmostEqual  # type: ignore

        else:
            comparator = self.assertEqual  # type: ignore

        comparator(  # type: ignore
            got,
            expected,
            msg=msg_format.format(
                input=repr(self),
                expected=repr(expected),
                output=repr(got),
                diff_explanation=diff_explanation,
                diff=diff,
            ),
        )

    def generate_test_case(
        self,
        golden: Callable[..., Output],
        under_test: Callable[..., Output],
        score: float,
        config: AgaConfig,
    ) -> AgaTestCase:
        """Generate a TestCase which tests `golden` against `under_test`."""
        metadata = TestMetadata(
            hidden=self.aga_kwargs.hidden,
            config=config.test,
            max_score=score,
            check_stdout=config.problem.check_stdout,
            mock_input=config.problem.mock_input,
        )
        return AgaTestCase(self, golden, under_test, metadata)

    def __repr__(self) -> str:
        """Get a string representation of the test case."""
        args_repr = self._param.args_repr(",")
        kwargs_repr = self._param.kwargs_repr(",")
        sep = self._param.sep_repr(",")

        return args_repr + sep + kwargs_repr

    def check_one(self, golden: Callable[..., Output]) -> None:
        """Check that the golden solution is correct."""
        if (
            self.aga_kwargs.expect is not None
            or self.aga_kwargs.expect_stdout is not None
        ):
            if self.aga_kwargs.override_test:

                def dummy_tester(*_: Any, **__: Any) -> Output:
                    # https://github.com/python/mypy/issues/4805 ehh
                    return self.aga_kwargs.expect  # type: ignore

                try:
                    self.aga_kwargs.override_test(
                        self,
                        dummy_tester,
                        golden,
                    )
                except AssertionError as e:
                    raise AssertionError(
                        "Your solution doesn't pass the overridden test. \n"
                        "Test parameters: \n"
                        f"{self._param}"
                    ) from e
            else:
                self._default_check(golden)

    def _default_check(self, golden: Callable[..., Output]) -> None:
        """Check that the golden solution is correct."""
        eval_fn = self._pick_eval_fn()
        golden_stdout, golden_output = eval_fn(golden, True)

        # compare output
        if self.aga_kwargs.expect is not None:
            compared_expect_out = self.aga_kwargs.expect
            if self.aga_kwargs.is_pipeline:
                golden_output = golden_output[1:]
                compared_expect_out = list(self.aga_kwargs.expect)

            # if check is overridden, use it
            if self.aga_kwargs.override_check is not None:
                try:
                    self.aga_kwargs.override_check(
                        self, compared_expect_out, golden_output
                    )
                except AssertionError as e:
                    raise AssertionError(
                        "Your solution doesn't pass the overridden check. \n"
                        "Test parameters: \n"
                        f"{self._param}"
                    ) from e
            else:
                self.assertEqual(golden_output, self.aga_kwargs.expect)

        # compare stdout
        if self.aga_kwargs.expect_stdout is not None:
            if isinstance(self.aga_kwargs.expect_stdout, str):
                compared_golden_stdout = golden_stdout
                compared_expected_stdout = self.aga_kwargs.expect_stdout
            elif isinstance(self.aga_kwargs.expect_stdout, Sequence) and isinstance(
                golden_stdout, str
            ):
                compared_golden_stdout = golden_stdout.splitlines()
                compared_expected_stdout = list(self.aga_kwargs.expect_stdout)
            else:
                raise TypeError(
                    "expect_stdout must be a string or a sequence of strings"
                )

            self.assertEqual(compared_expected_stdout, compared_golden_stdout)


class _TestInputGroup(Generic[Output]):
    """A group of test cases with shared configuration."""

    def __init__(
        self, weight: int = 1, value: float = 0.0, extra_credit: float = 0.0
    ) -> None:
        self._test_cases: list[_TestInputs[Output]] = []
        self._prizes: list[Prize] = []
        self.score_info = ScoreInfo(weight, value, extra_credit)

    def add_test_case(self, case: _TestInputs[Output]) -> None:
        """Add a test case to the group."""
        self._test_cases.append(case)

    def add_prize(self, prize: Prize) -> None:
        """Add a prize to the group."""
        self._prizes.append(prize)

    def generate_test_suite(
        self,
        golden: Callable[..., Output],
        under_test: Callable[..., Output],
        group_score: float,
        config: AgaConfig,
    ) -> tuple[AgaTestSuite, list[ScoredPrize]]:
        """Generate a test suite from all the test cases for this group."""
        suite = AgaTestSuite(config, [])

        score_infos = [case.score_info for case in self._test_cases] + [
            prize.score_info for prize in self._prizes
        ]
        scores = compute_scores(score_infos, group_score)

        for score, case in zip(
            scores, self._test_cases
        ):  # type: float, _TestInputs[Output]
            suite.addTest(case.generate_test_case(golden, under_test, score, config))

        scored_prizes = []
        for score, prize in zip(reversed(scores), reversed(self._prizes)):
            # reverse scores so we get the ones that correspond to the prizes
            scored_prizes.append(
                ScoredPrize(prize=prize, max_score=score)  # type: ignore
            )

        return suite, scored_prizes

    def check_one(self, golden: Callable[..., Output]) -> None:
        """Check the golden solution against all test cases."""
        for case in self._test_cases:
            case.check_one(golden)



"""Utility classes for the core module."""

from __future__ import annotations

from contextlib import redirect_stdout
from io import StringIO
from typing import Any, Type, Sequence, Dict

__all__ = (
    "CaptureOut",
    "MethodCaller",
    "MethodCallerFactory",
    "PropertyGetter",
    "PropertyGetterFactory",
    "Initializer",
    "initializer",
)


def _ensure_formatting_dot(s: str) -> str:
    """Ensure that a string ends with a dot."""
    if s[0] != ".":
        return "." + s
    return s


class CaptureOut:
    """Context manager for capturing stdout."""

    def __init__(self, capture: bool):
        """Initialize the context manager."""
        self.capture: bool = capture
        self.capture_device: redirect_stdout[StringIO] | None = None
        self.io_device: StringIO | None = None

    def __enter__(self) -> CaptureOut:
        """Enter the context manager."""
        if self.capture:
            self.io_device = StringIO()
            self.capture_device = redirect_stdout(self.io_device)
            self.capture_device.__enter__()
        return self

    def __exit__(self, *args: Any) -> Any:
        """Exit the context manager."""
        if self.capture and self.capture_device:
            return self.capture_device.__exit__(*args)
        return None

    @property
    def value(self) -> str | None:
        """Return the captured output, or None if not capturing."""
        if self.capture and self.io_device:
            return self.io_device.getvalue()
        else:
            return None


# pylint: disable=too-few-public-methods
class MethodCaller:
    """Call a method on an instance."""

    def __init__(self, attr_name: str, *args: Any, **kwargs: Any):
        self._attr_name = attr_name
        self._args = args
        self._kwargs = kwargs

    @property
    def attr_name(self) -> str:
        """Return the method attribute name."""
        return self._attr_name

    @property
    def args(self) -> Sequence[Any]:
        """Return the positional arguments."""
        return self._args

    @property
    def kwargs(self) -> Dict[str, Any]:
        """Return the keyword arguments."""
        return self._kwargs

    def __call__(self, instance: Any, previous_result: Any = None) -> Any:
        """Call the method on the instance."""
        return getattr(instance, self.attr_name)(*self.args, **self.kwargs)

    def __str__(self) -> str:
        """Return a string representation of the method call."""
        arg_repr = f'{", ".join(map(repr, self.args))}, ' if self.args else ""
        kwargs = (f"{k}={repr(v)}" for k, v in self.kwargs.items())  # type: ignore
        kwargs_repr = f'{", ".join(kwargs)}, ' if self.kwargs else ""
        return _ensure_formatting_dot(f"{self.attr_name}({arg_repr}{kwargs_repr})")

    def __repr__(self) -> str:
        """Return a string representation of the method call."""
        return self.__str__()


class MethodCallerFactory:
    """Factory for creating MethodCaller instances."""

    def __init__(self, attr_name: str):
        self._attr_name = attr_name

    @property
    def attr_name(self) -> str:
        """Return the method attribute name."""
        return self._attr_name

    def __call__(
        self, *args: Any, caller_class: Type[MethodCaller] = MethodCaller, **kwargs: Any
    ) -> MethodCaller:
        """Create a MethodCaller instance with the method name."""
        return caller_class(self._attr_name, *args, **kwargs)


# pylint: disable=too-few-public-methods
class PropertyGetter:
    """Get an attribute from an instance."""

    def __init__(self, *args: str) -> None:
        self._attr_names = args

    @property
    def attr_names(self) -> Sequence[str]:
        """Return the attribute name."""
        return self._attr_names

    def __str__(self) -> str:
        """Return a string representation of the property getter."""
        return _ensure_formatting_dot(
            ".".join(
                name[1:] if name.startswith(".") else name for name in self.attr_names
            )
        )

    def __repr__(self) -> str:
        """Return a string representation of the property getter."""
        return self.__str__()

    def __call__(self, instance: Any, previous_result: Any = None) -> Any:
        """Get the attribute from the instance."""
        if len(self.attr_names) == 0:
            raise ValueError("No properties specified in PropertyGetter")

        for i, attr_name in enumerate(self.attr_names):
            properties = attr_name.split(".")

            if properties[0] == "":
                properties = properties[1:]

            if len(properties) == 0:
                raise ValueError(
                    f"No properties specified in the {i}th attribute "
                    f"name in the PropertyGetter. "
                    f"All attributes specified: {self.attr_names}"
                )

            for property_name in properties:
                if property_name == "":
                    raise ValueError(f"Double . in {self.attr_names}")
                instance = getattr(instance, property_name)

        return instance


class PropertyGetterFactory:
    """Factory for creating PropertyGetter instances."""

    def __call__(
        self,
        *args: Any,
        getter_class: Type[PropertyGetter] = PropertyGetter,
        **kwargs: Any,
    ) -> PropertyGetter:
        """Create a PropertyGetter instance with the attribute name."""
        return getter_class(*args, **kwargs)


# pylint: disable=too-few-public-methods
class Initializer(MethodCaller):
    """Call the __init__ method on an instance."""

    DEFAULT_NEW_INSTANCE_NAME = "instance"

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__("__call__", *args, **kwargs)


initializer = Initializer()
